package ru.dbotthepony.kstarbound.world

import ru.dbotthepony.kstarbound.RegistryObject
import ru.dbotthepony.kstarbound.defs.tile.LiquidDefinition
import ru.dbotthepony.kstarbound.defs.tile.MaterialModifier
import ru.dbotthepony.kstarbound.defs.tile.TileDefinition
import java.io.DataInputStream

enum class TileColor {
	DEFAULT,
	RED,
	BLUE,
	GREEN,
	YELLOW,
	BROWN,
	PURPLE,
	BLACK,
	WHITE;

	companion object {
		private val values = values()

		fun of(index: Int): TileColor {
			return values.getOrElse(index) { DEFAULT }
		}
	}
}

interface ITileState {
	var material: TileDefinition?
	var modifier: MaterialModifier?
	var color: TileColor
	var hueShift: Float
	var modifierHueShift: Float

	val isEmpty: Boolean

	fun setHueShift(value: Int) {
		if (value < 0) {
			hueShift = 0f
		} else if (value >= 255) {
			hueShift = 360f
		} else {
			hueShift = value / 255f * 360f
		}
	}

	fun setModHueShift(value: Int) {
		if (value < 0) {
			modifierHueShift = 0f
		} else if (value >= 255) {
			modifierHueShift = 360f
		} else {
			modifierHueShift = value / 255f * 360f
		}
	}

	fun read(
		materialAccess: (Int) -> RegistryObject<TileDefinition>?,
		modifierAccess: (Int) -> RegistryObject<MaterialModifier>?,
		stream: DataInputStream
	) {
		material = materialAccess(stream.readUnsignedShort())?.value
		setHueShift(stream.read())
		color = TileColor.of(stream.read())
		modifier = modifierAccess(stream.readUnsignedShort())?.value
		setModHueShift(stream.read())
	}

	companion object : ITileState {
		override var material: TileDefinition?
			get() = null
			set(value) { throw UnsupportedOperationException("Empty tile state") }
		override var modifier: MaterialModifier?
			get() = null
			set(value) { throw UnsupportedOperationException("Empty tile state") }
		override var color: TileColor
			get() = TileColor.DEFAULT
			set(value) { throw UnsupportedOperationException("Empty tile state") }
		override var hueShift: Float
			get() = 0f
			set(value) { throw UnsupportedOperationException("Empty tile state") }
		override var modifierHueShift: Float
			get() = 0f
			set(value) { throw UnsupportedOperationException("Empty tile state") }
		override val isEmpty: Boolean
			get() = true
	}
}

interface ILiquidState {
	var def: LiquidDefinition?
	var level: Float
	var pressure: Float
	var isInfinite: Boolean

	val isEmpty: Boolean

	fun read(
		liquidAccess: (Int) -> RegistryObject<LiquidDefinition>?,
		stream: DataInputStream
	) {
		def = liquidAccess(stream.readUnsignedByte())?.value
		level = stream.readFloat()
		pressure = stream.readFloat()
		isInfinite = stream.readBoolean()
	}

	companion object : ILiquidState {
		override var def: LiquidDefinition?
			get() = null
			set(value) { throw UnsupportedOperationException("Empty liquid state") }
		override var level: Float
			get() = 0f
			set(value) { throw UnsupportedOperationException("Empty liquid state") }
		override var pressure: Float
			get() = 0f
			set(value) { throw UnsupportedOperationException("Empty liquid state") }
		override var isInfinite: Boolean
			get() = false
			set(value) { throw UnsupportedOperationException("Empty liquid state") }
		override val isEmpty: Boolean
			get() = true
	}
}

interface IChunkCell {
	val foreground: ITileState
	val background: ITileState
	val liquid: ILiquidState

	var dungeonId: Int
	var biome: Int
	var envBiome: Int
	var isIndestructible: Boolean

	val isEmpty: Boolean

	fun read(
		materialAccess: (Int) -> RegistryObject<TileDefinition>?,
		modifierAccess: (Int) -> RegistryObject<MaterialModifier>?,
		liquidAccess: (Int) -> RegistryObject<LiquidDefinition>?,
		stream: DataInputStream
	) {
		foreground.read(materialAccess, modifierAccess, stream)
		background.read(materialAccess, modifierAccess, stream)
		liquid.read(liquidAccess, stream)

		stream.skipBytes(1) // collisionMap

		dungeonId = stream.readUnsignedShort()
		biome = stream.readUnsignedByte()
		envBiome = stream.readUnsignedByte()
		isIndestructible = stream.readBoolean()

		stream.skipBytes(1) // unknown
	}

	companion object : IChunkCell {
		override val foreground: ITileState
			get() = ITileState.Companion
		override val background: ITileState
			get() = ITileState.Companion
		override val liquid: ILiquidState
			get() = ILiquidState.Companion
		override var dungeonId: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var biome: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var envBiome: Int
			get() = 0
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override var isIndestructible: Boolean
			get() = false
			set(value) { throw UnsupportedOperationException("Empty chunk cell") }
		override val isEmpty: Boolean
			get() = true
	}
}
